/* Copyright (c) 2005 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you
       must not claim that you wrote the original software. If you use
       this software in a product, an acknowledgment in the product
       documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and
       must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
       distribution. */

// $Header: /home/doj/code/doj/fft.hpp,v 1.5 2005-08-04 22:20:32 doj Exp $

#ifndef FFT__HPP
#define FFT__HPP

#include <fftw3.h>
#include <vector>
#include "windowfunctions.hpp"

namespace doj
{

  template <typename T, typename WF>
  std::vector<T> powerspectrum(T *samples, const unsigned s, WF window_func=doj::rectangular_window)
  {
    std::vector<T> ret;

    if(!samples || !s) return ret;

    double *in=reinterpret_cast<double*>(fftw_malloc(sizeof(double) * s));
    double *out=reinterpret_cast<double*>(fftw_malloc(sizeof(double) * s));
    fftw_plan plan=fftw_plan_r2r_1d(s, in, out, FFTW_R2HC, /*FFTW_ESTIMATE|*/FFTW_DESTROY_INPUT);

    for(unsigned i=0; i!=s; ++i)
      in[i]=static_cast<double>(samples[i]) * window_func(i,s);

    fftw_execute(plan);

    const unsigned ke=(s+1)/2;
    for(unsigned k=1; k<ke; ++k)
      in[k] = ::sqrt(out[k]*out[k] + out[s-k]*out[s-k]);
    if(s%2 == 0)
      in[ke] = ::sqrt(out[ke]*out[ke]); // Nyquist freq.

    fftw_destroy_plan(plan);
    fftw_free(out);

    ret.reserve(ke);
    for(unsigned i=0; i!=ke; ++i)
      ret.push_back(static_cast<T>(in[i]));

    fftw_free(in);

    return ret;
  }

  template <typename T, typename WF>
  std::vector<T> powerspectrum(std::vector<T> samples, WF window_func=doj::rectangular_window)
  {
    return powerspectrum(&samples[0], samples.size(), window_func);
  }

  inline double fft_bin2freq(const int binnr, const int fftsize, const int samplerate)
  {
    const int sr=samplerate/2;
    if(sr<1)
      return 0;
    if(fftsize==0)
      return 0;
    const int ff=fftsize/2;
    if(binnr>ff)
      return sr;
    if(binnr<1)
      return 0;
    return (static_cast<double>(binnr)-.5) / static_cast<double>(ff) * static_cast<double>(sr);
  }

  inline int fft_freq2bin(const double freq, const int fftsize, const int samplerate)
  {
    const int sr=samplerate/2;
    if(sr<1)
      return 0;
    if(static_cast<int>(freq) >= sr)
      return fftsize/2;
    return static_cast<int>(freq/static_cast<double>(sr)*static_cast<double>(fftsize/2) +.5);
  }

}

#endif
