/* The zlib/libpng License

Copyright (c) 2005 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
   not claim that you wrote the original software. If you use this
   software in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
   distribution.
*/

#include "sndfile.hpp"
#include "fft.hpp"
#include <algorithm>
#include <assert.h>
#include <iomanip>
#include <sstream>
#include <unistd.h>
using namespace doj;
using namespace std;

#include "common.cpp"

void help()
{
  cerr << "usage: sample2sonogram [-2..-9] [-b] [-f framesize] [-h height] [-l] [-o offset] [-p] [-s fontsize] [-t text] [-w width] [-z] <filename>\n"
       << "  -2..-9 show only lower half/quarter/eigth of sonogram\n"
       << "  -b draw box around image\n"
       << "  -f framesize in samples\n"
       << "  -h height in cm\n"
       << "  -l draw lines\n"
       << "  -o offset in cm from lower left corner\n"
       << "  -p for printing optimized greyscale palette\n"
       << "  -s fontsize in pt\n"
       << "  -t print text in lower left corner. if text is a single space character the filename and samplerate are printed\n"
       << "  -w width in cm\n"
       << "  -z draw sample\n"
       << "version " << VERSION << "\n"
       << "convered by the zlib/libpng license\n"
       << "(c) by Dirk Jagdmann <doj@cubic.org>\n"
       << "http://llg.cubic.org/tools/sample2ps\n"
    ;

  exit(1);
}

int main(int argc, char **argv)
{
  float width=cm2pt(18), height=-1, offset=0;
  bool drawbox=false, forprint=false, drawlines=false, drawsample=false;
  string text;
  int framesize=4096, scale=1;
  int c;
  while((c=getopt(argc, argv, "123456789bf:h:lo:ps:t:w:z")) >= 0)
    switch(c)
      {
      default: help();

      case '1': scale=1; break;
      case '2': scale=2; break;
      case '3': scale=3; break;
      case '4': scale=4; break;
      case '5': scale=5; break;
      case '6': scale=6; break;
      case '7': scale=7; break;
      case '8': scale=8; break;
      case '9': scale=9; break;

      case 'b': drawbox=true; break;

      case 'f':
	framesize=strtol(optarg,0,0);
	if(framesize<2 && (framesize&1))
	  help();
	break;

      case 'h':
	height=cm2pt(strtof(optarg,0));
	if(height<=0.0)
	  help();
	break;

      case 'l': drawlines=true; break;

      case 'o': offset=cm2pt(strtof(optarg,0)); break;

      case 'p': forprint=true; break;

      case 's':
	fontsize=strtof(optarg,0);
	if(fontsize<=0.0)
	  help();
	break;

      case 't': text=optarg; break;

      case 'w':
	width=cm2pt(strtof(optarg,0));
	if(width<=0.0)
	  help();
	break;

      case 'z': drawsample=true; break;
      }

  if(height<0.0)
    height=width/1.6;

  if(optind==argc)
    help();

  // load sample
  const string filename=argv[optind];

  int samplerate, channels;
  unsigned size;
  double *buf=0;
  buf=sndfile_read(filename, samplerate, channels, size, buf);
  if(!buf)
    return 1;
  if(channels!=1)
    {
      cerr << "channels != 1\n";
      return 1;
    }
  if(!size)
    {
      cerr << "size==0\n";
      return 1;
    }

  // setup float format
  cout << fixed << setprecision(2);

  // output header
  cout << "%!PS-Adobe-3.0 EPSF-3.0\n"
       << "%%BoundingBox: " << ((offset!=0.0)?static_cast<int>(offset-1.0):0) << " "
                            << ((offset!=0.0)?static_cast<int>(offset-1.0):0) << " "
                            << static_cast<int>(width+offset+2.0) << " "
                            << static_cast<int>(height+offset+2.0) << "\n"
       << "%%Title: " << argv[optind] << "\n"
       << "%%Creator: sample2sonogram http://llg.cubic.org/tools/sample2ps\n"
       << "%%Palette: " << (forprint?"print optimized":"normal") << "\n"
       << "gsave\n"

    // palette function
       << " /b_1_3 6 array def\n"
       << " /g_1_2 6 array def\n"
       << " /r_1_1 6 array def\n"
       << " /i { % Set new frame with 4 slots\n"
       << " % and intitialize variables to 0\n"
       << " 4 dict begin\n"
       << " /x_2_3 0 def\n"
       << " /y_2_4 0 def\n"
       << " /f_2_2 0 def\n"
       << " /c_2_1 exch def\n"
       << " /f_2_2  size_1_4  1  sub\n"
       << " store\n"
       << " /f_2_2  1  f_2_2  div\n"
       << " store\n"
       << " /x_2_3  1  store\n"
       << "{ % beginning of while loop\n"
       << " x_2_3  size_1_4  1  sub\n"
       << " lt  c_2_1  f_2_2  gt  and\n"
       << "\n"
       << "  {  /x_2_3  x_2_3  1  add\n"
       << " store\n"
       << " /c_2_1  c_2_1  f_2_2  sub\n"
       << " store\n"
       << "  }\n"
       << " {exit} ifelse\n"
       << "} loop %  end of while\n"
       << " /y_2_4  x_2_3  1  add\n"
       << " store\n"
       << " /c_2_1  c_2_1  f_2_2  div\n"
       << " store\n"
       << " r_1_1  x_2_3 1 sub\n"
       << " get  1  c_2_1  sub\n"
       << " mul\n"
       << " r_1_1  y_2_4 1 sub\n"
       << " get  c_2_1  mul\n"
       << " add\n"
       << " g_1_2  x_2_3 1 sub\n"
       << " get  1  c_2_1  sub\n"
       << " mul\n"
       << " g_1_2  y_2_4 1 sub\n"
       << " get  c_2_1  mul\n"
       << " add\n"
       << " b_1_3  x_2_3 1 sub\n"
       << " get  1  c_2_1  sub\n"
       << " mul\n"
       << " b_1_3  y_2_4 1 sub\n"
       << " get  c_2_1  mul\n"
       << " add\n"
       << "setrgbcolor\n"
       << " r_1_1  g_1_2  b_1_3  size_1_4  end\n"
       << " /size_1_4 exch def\n"
       << " /b_1_3 exch def\n"
       << " /g_1_2 exch def\n"
       << " /r_1_1 exch def\n"
       << "  } store\n"
       << " /size_1_4  6  store\n"
       << " r_1_1  1 1 sub\n"
       << " 1  put\n"
       << " g_1_2  1 1 sub\n"
       << " 0  put\n"
       << " b_1_3  1 1 sub\n"
       << " 0.25  put\n"
       << " r_1_1  2 1 sub\n"
       << " 1  put\n"
       << " g_1_2  2 1 sub\n"
       << " 0  put\n"
       << " b_1_3  2 1 sub\n"
       << " 0  put\n"
       << " r_1_1  3 1 sub\n"
       << " 1  put\n"
       << " g_1_2  3 1 sub\n"
       << " 1  put\n"
       << " b_1_3  3 1 sub\n"
       << " 0  put\n"
       << " r_1_1  4 1 sub\n"
       << " 1  put\n"
       << " g_1_2  4 1 sub\n"
       << " 1  put\n"
       << " b_1_3  4 1 sub\n"
       << " 0  put\n"
       << " r_1_1  5 1 sub\n"
       << " 0  put\n"
       << " g_1_2  5 1 sub\n"
       << " 0  put\n"
       << " b_1_3  5 1 sub\n"
       << " 1  put\n"
       << " r_1_1  6 1 sub\n"
       << " 0  put\n"
       << " g_1_2  6 1 sub\n"
       << " 0  put\n"
       << " b_1_3  6 1 sub\n"
       << " 0.2  put\n"

    // translate one up
       << "/t { 0 1 translate } def\n"

    // generic box
       << "/bx { "
       << "  newpath 0 0 moveto 0 1 rlineto 1 0 rlineto 0 -1 rlineto closepath "
       << "  fill "
       << "  t "
       << "} def\n"

    // color box
       << "/bc { i bx } def\n"

    // gray box
       << "/bg { setgray bx } def\n"

    // outer frame
       << "/fr { "
       << "  newpath "
       << "  0 0 moveto "
       << "  0 " << height << " lineto "
       << width << " " << height << " lineto "
       << width << " 0 lineto "
       << "  closepath "
       << "} def\n"

    // look to advance in y
       << "/r { {t} repeat } def\n";
    ;

  if(offset!=0.0)
    cout << offset << " dup translate\n";

  // prepare font
  rgba fontcol=forprint?rgba(0,0,0):rgba(1,1,1);
  setupFont("Times-Bold");

  // draw background
  if(!forprint)
    {
      setcol(0.0) << " fr fill\n";
    }

  // draw lines
  if(drawlines)
    {
      cout << "gsave\n";
      cout << "0.25 setlinewidth\n";
      rgba col=forprint?rgba(.2,.2,.2):rgba(0,.5,0);
      int Y=static_cast<int>(pt2cm(height));
      if(Y&1) ++Y;		// make Y even
      for(int y=1; y<Y; ++y)
	{
	  const double fac=static_cast<double>(y)/Y;
	  const double A=height*fac;

	  setcol(col) << " newpath 0 " << A << " moveto " << width << " " << A << " lineto stroke\n";

	  const double sr=fac*samplerate*.5/scale;
	  ostringstream o; o << static_cast<int>(sr) << "Hz";
	  print(o.str(), 0.25*fontsize, A, fontcol, Left|Mid);

	}
      cout << "grestore\n";
    }

#if 1
  // calculate sonogram
  typedef vector<double> v_t;
  v_t son; son.reserve(size);
  for(unsigned i=0; i<size-framesize; i+=framesize/2)
    {
      v_t s=powerspectrum(buf+i, framesize, blackman_window);
      copy(s.begin(), s.end(), back_inserter(son));
    }
  v_t::value_type max=0;
  for(v_t::iterator i=son.begin(); i!=son.end(); ++i)
    if(*i>max)
      max=*i;
  if(max!=0.0)
    {
      max=1.0/log(max+1.0);
      for(v_t::iterator i=son.begin(); i!=son.end(); ++i)
	{
	  if(*i < 0.0) *i=0;
	  *i = 1.0 - log(*i+1.0)*max;
	  assert(finite(*i));
	}
    }

  cout << "gsave\n"
       << width/(son.size()/(framesize/2)) << " " << height/(framesize/2) << " scale\n"
       << "0 0 moveto\n"
    ;

  if(!forprint)
    highthreshold+=0.05;

  unsigned i=0;
  while(i<son.size())
    {
      cout << "gsave\n";

      if(scale!=1)
	cout << "1 " << scale << " scale\n";

      // search for highest freq
      int high=framesize/2/scale-1;
      for(; high>0; --high)
	if(son[i+high]<highthreshold)
	  break;

      // plot until highes freq
      ++i; // ignore DC
      int y, skip=0;
      for(y=1; y<=high; ++y, ++i)
	{
	  // if value too bright, don't output a box
	  if(son[i]>highthreshold)
	    {
#define DEB 0
#if !DEB
	      ++skip;
#else
	      print(1) << " bc\n"; // for debugging
#endif
	    }
	  else
	    {
#if !DEB
	      if(skip)
		{
		  if(skip==1)
		    cout << "t\n";
		  else
		    cout << skip << " r\n";
		  skip=0;
		}
#endif
	      print(son[i]) << (forprint?" bg\n":" bc\n");
	    }
	}
      i+=framesize/2-y;
#if DEB
      print(rgba(1,0,0)) << " bc\n";// for debugging, print a red box
#endif
#undef DEB
      cout << "grestore\n";
      cout << "1 0 translate\n";
    }
  // draw a white box
  const int ysize=framesize/2/scale;
  cout << "-2 " << 0.95*ysize << " translate 1 bg\n";
  cout << "grestore\n";

  const int xsize=i/(framesize/2);
  cout << "% Dimension " << xsize << "x" << ysize << endl;
#endif

  // draw sample
  if(drawsample)
    {
      const unsigned W=2048;
      const unsigned step=size/W;
      const double amp=height*.5;
#define Y(x) x*amp+amp
      cout << "gsave\n"
	   << width/W << " 1 scale\n"
	   << "1 setlinecap 1 setlinejoin\n"
	   << "0.2 setlinewidth\n"
	   << "0.2 .6 .6 setrgbcolor\n"
	   << "newpath\n"
	;
      bool first=true;
      for(unsigned x=0, z=0; x<W; ++x)
	{
	  double acc=0;
	  for(unsigned i=0; i!=step; ++i)
	    acc+=buf[z++];
	  acc/=step;
	  if(std::fabs(acc)<0.001) acc=0;

	  if(first && acc!=0.0)
	    {
	      cout << x << " " << Y(acc) << " moveto\n";
	      first=false;
	    }
	  else if(!first && acc==0.0)
	    {
	      cout << x << " " << Y(0) << " lineto stroke\n";
	      first=true;
	    }
	  else if(!first)
	    cout << x << " " << Y(acc) << " lineto\n";
	}
      cout << "stroke\n"
	   << "grestore\n"
	;
#undef Y
    }

  // print texts
  if(text.size())
    {
      string t;
      if(text==" ")
	{
	  ostringstream o;
	  o << filename << " " << samplerate << "Hz " << static_cast<double>(size)/samplerate << "s";
	  t=o.str();
	}
      else
	t=text;

      print(t, 0.25*fontsize, 0.25*fontsize, fontcol);
    }

  // draw box
  if(drawbox)
    setcol(0.0) << "fr 1 setlinewidth stroke\n";

  cout << "grestore\n";
  cout << "showpage\n";
  return 0;
}
