/*

Formulas and descriptions were gathered (ripped) from various sources, including:
http://www.mathworks.com

*/

#ifndef WINDOWFUNCTIONS__HPP
#define WINDOWFUNCTIONS__HPP

#include "constant.hpp"

namespace doj
{

  inline double rectangular_window(double pos, double len)
  {
    return 1.0;
  }

  /**
     calculate a triangle window. Also known as Bartlett window. The Frequency response is sinc^2.
     @param pos current position in window
     @param len window length
  */
  inline double triangle_window(double pos, double len)
  {
    const double l=len/2.0;
    return (pos<=l) ? pos/l : (len-pos)/l;
  }

  /**
     calculate a Hanning window
     @param pos current position in window
     @param len window length
  */
  inline double hanning_window(double pos, double len)
  {
    return 0.5 - 0.5*cos(2.0*M_PI*pos/len);
  }

  /**
     calculate a Hamming window
     @param pos current position in window
     @param len window length
  */
  inline double hamming_window(double pos, double len)
  {
    const double a=25.0/46.0;
    return /*0.54*/a - /*0.46*/(1.0-a)*cos(2.0*M_PI*pos/len);
  }

  /**
     Blackman windows have slightly wider central lobes and less
     sideband leakage than equivalent length Hamming and Hann windows.

     @param pos current position in window
     @param len window length
  */
  inline double blackman_window(double pos, double len)
  {
    return 0.42 - 0.5*cos(2.0*M_PI*pos/len) + 0.08*cos(4.0*M_PI*pos/len);
  }

  inline double generic_blackman_harris_4term_window(double pos, double len, double a0, double a1, double a2, double a3)
  {
    const double t=pos/len;
    return a0 + a1*cos(2.0*M_PI*t) + a2*cos(4.0*M_PI*t) + a3*cos(6.0*M_PI*t);

  }

  /**
     The window is minimum in the sense that its maximum sidelobes are
     minimized.
     @param pos current position in window
     @param len window length
  */
  inline double blackman_harris_window(double pos, double len)
  {
    return generic_blackman_harris_4term_window(pos, len, 0.35875, -0.48829, 0.14128, -0.01168);
  }

  /**
     The window is minimum in the sense that its maximum sidelobes are
     minimized. The coefficients for this window differ from the
     Blackman-Harris window coefficients computed with blackman_harris_window()
     and produce slightly lower sidelobes.
     @param pos current position in window
     @param len window length
  */
  inline double nuttall_window(double pos, double len)
  {
    return generic_blackman_harris_4term_window(pos, len, 0.3635819, -0.4891775, 0.1365995, -0.0106411);
  }

  /**
     A Bohman window is the convolution of two half-duration cosine
     lobes. In the time domain, it is the product of a triangular
     window and a single cycle of a cosine with a term added to set
     the first derivative to zero at the boundary. Bohman windows fall
     off as 1/w^4.
     @param pos current position in window
     @param len window length
  */
  inline double bohman_window(double pos, double len)
  {
    const double len2=len*.5;
    return (1.0-(pos-len2)/len2)*cos(M_PI*(pos-len2)/len2) + M_1_PI*sin(M_PI*(pos-len2)/len2);
  }

  /**
     Flat Top windows have very low passband ripple (< 0.01 dB) and
     are used primarily for calibration purposes. Their bandwidth is
     approximately 2.5 times wider than a Hanning window.
     @param pos current position in window
     @param len window length
  */
  inline double flattop_window(double pos, double len)
  {
    const double t=pos/len;
    return 1.0 - 1.93*cos(2.0*M_PI*t) + 1.29*cos(4.0*M_PI*t) - 0.388*cos(6.0*M_PI*t) + 0.0322*cos(8.0*M_PI*t);
  }

  /**
     @param pos current position in window
     @param len window length
     @param alpha the reciprocal of the standard deviation. A larger value of alpha produces a more narrow window.
  */
  inline double gauss_window(double pos, double len, double alpha=2.5)
  {
    const double len2=len*.5;
    const double r=alpha*(pos-len2)/len2;
    return exp( -.5 * r*r );
  }

  /**
     Parzen (de la Valle-Poussin) windows are piecewise cubic
     approximations of Gaussian windows. Parzen window sidelobes fall
     off as 1/w^4.
     @param pos current position in window
     @param len window length
  */
  inline double parzen_window(double pos, double len)
  {
    const double len2=len*.5;
    const double abspos=fabs(pos);
    if(abspos <= len2*.5)
      {
	const double a=pos/len2;
	return 1.0 - 6.0*a*a*(1.0 - abspos/len2);
      }
    const double a=1.0 - abspos/len2;
    return 2.0 * a*a*a;
  }

  /**
     Tukey windows are cosine-tapered windows.

     @param pos current position in window
     @param len window length
     @param r between 0 and 1. 0:=rectangular window, 1:=hanning window
  */
  inline double tukey_window(double pos, double len, double r)
  {
    const double R=r*.5*len;
    if(pos < R)
      return .5 * (1.0 + cos(2.0*M_PI/r * pos/len - M_PI));
    if(pos < len - R)
      return 1.0;
    return .5 * (1.0 + cos(2.0*M_PI/r - 2.0*M_PI/r*pos/len - M_PI));
  }

  /**
     Like Bartlett, Hann, and Hamming windows, this window has a
     mainlobe at the origin and asymptotically decaying sidelobes on
     both sides. It is a linear combination of weighted Bartlett and
     Hann windows with near sidelobes lower than both Bartlett and
     Hann and with far sidelobes lower than both Bartlett and Hamming
     windows. The mainlobe width of the modified Bartlett-Hann window
     is not increased relative to either Bartlett or Hann window
     mainlobes.
     @param pos current position in window
     @param len window length
  */
  inline double modified_bartlett_hanning_window(double pos, double len)
  {
    return 0.62 - 0.48*fabs(pos/len - 0.5) + 0.38*cos(2.0*M_PI*(pos/len - 0.5));
  }

#if 0
  /**
     Chebyshev window whose Fourier transform sidelobe magnitude is r
     dB below the mainlobe magnitude. The default value for r is 100.0
     dB.
  */
  inline double chebyshev_window(double pos, double len, double r)
  {

  }

  inline double kaiser_window(double pos, double len)
  {

  }
#endif

  /**
     @param pos current position in window
     @param len window length
  */
  inline double welch_window(double pos, double len)
  {
    const double a=(pos-len*.5)/(len*.5);
    return 1.0 - a*a;
  }

}

#endif
